/**
 * WEBLAB: Service oriented integration platform for media mining and intelligence applications
 * 
 * Copyright (C) 2004 - 2010 EADS - CASSIDIAN
 * 
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License along with this
 * library; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA 02110-1301 USA
 */

package org.ow2.weblab.portlet;

import java.io.IOException;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import javax.portlet.ActionRequest;
import javax.portlet.ActionResponse;
import javax.portlet.EventRequest;
import javax.portlet.EventResponse;
import javax.portlet.PortletConfig;
import javax.portlet.PortletException;
import javax.portlet.PortletRequestDispatcher;
import javax.portlet.PortletSession;
import javax.portlet.RenderRequest;
import javax.portlet.RenderResponse;
import javax.xml.namespace.QName;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.ow2.weblab.core.extended.exception.WebLabCheckedException;
import org.ow2.weblab.core.extended.ontologies.WebLabRetrieval;
import org.ow2.weblab.core.extended.util.ResourceUtil;
import org.ow2.weblab.core.helper.impl.JenaPoKHelper;
import org.ow2.weblab.core.helper.impl.JenaResourceHelper;
import org.ow2.weblab.core.model.Annotation;
import org.ow2.weblab.core.model.ComposedQuery;
import org.ow2.weblab.core.model.Query;
import org.ow2.weblab.core.model.Resource;
import org.ow2.weblab.core.model.ResultSet;
import org.ow2.weblab.core.model.StringQuery;
import org.ow2.weblab.core.services.searcher.SearchArgs;
import org.ow2.weblab.portlet.bean.Facet;
import org.ow2.weblab.portlet.bean.FacetBundle;
import org.ow2.weblab.portlet.bean.FacetCategory;
import org.ow2.weblab.portlet.conf.FacetConfig;
import org.springframework.beans.factory.BeanFactory;
import org.springframework.beans.factory.xml.XmlBeanFactory;
import org.springframework.core.io.ClassPathResource;

import com.hp.hpl.jena.rdf.model.Model;
import com.hp.hpl.jena.rdf.model.Statement;

public class FacetPortlet extends WebLabPortlet {

	// TODO these 4 properties will be part of model 1.2.5.
	// Use annotator when this is ready.
	public static final String TO_BE_RANKED_BY = "http://weblab.ow2.org/core/1.2/ontology/retrieval#toBeRankedBy";
	public static final String TO_BE_RANKED_ASCENDING = "http://weblab.ow2.org/core/1.2/ontology/retrieval#toBeRankedAscending";
	public static final String HAS_SCOPE = "http://weblab.ow2.org/core/1.2/ontology/retrieval#hasScope";
	public static final String IS_RANKED_ASCENDING = "http://weblab.ow2.org/core/1.2/ontology/retrieval#isRankedAscending";

	public static final String FACET_SUGGESTION = "FacetSuggestion";

	private final static Log logger = LogFactory.getLog(FacetPortlet.class);

	private FacetConfig portletConf;

	protected PortletRequestDispatcher normalView;
	protected PortletRequestDispatcher noDataView;
	protected PortletRequestDispatcher errorView;

	@Override
	public void init(PortletConfig config) throws PortletException {
		super.init(config);
		super.init();
		normalView = config.getPortletContext().getRequestDispatcher(getInitParameter(FacetPortletConstants.NORMAL_PAGE));
		noDataView = config.getPortletContext().getRequestDispatcher(getInitParameter(FacetPortletConstants.NO_DATA_PAGE));
		errorView = config.getPortletContext().getRequestDispatcher(getInitParameter(FacetPortletConstants.ERROR_PAGE));

		ClassPathResource resource = new ClassPathResource(FacetConfig.DEFAULT_CONF_FILE);
		BeanFactory factory = new XmlBeanFactory(resource);
		this.portletConf = factory.getBean(FacetConfig.DEFAULT_BEAN_NAME, FacetConfig.class);
	}

	@Override
	protected void doView(RenderRequest req, RenderResponse res) throws PortletException, IOException {

		if (req.getParameter(FacetPortletConstants.FACET_ORDER) == null) {
			req.getPortletSession().setAttribute(FacetPortletConstants.FACET_ORDER, portletConf.getFacetsList(), PortletSession.APPLICATION_SCOPE);
		}

		if (req.getParameter(FacetPortletConstants.ERROR) != null) {
			// dispatch to FacetPortletConstants.ERROR jsp
			errorView.include(req, res);
		}

		if (req.getPortletSession().getAttribute(FacetPortletConstants.FACET_CATS, PortletSession.APPLICATION_SCOPE) == null) {
			// dispatch to no data jsp
			noDataView.include(req, res);
		} else {
			req.setAttribute(FacetPortletConstants.FACET_ORDER, req.getParameter(FacetPortletConstants.FACET_ORDER));
			req.setAttribute(FacetPortletConstants.FACET_CATS,
					req.getPortletSession().getAttribute(FacetPortletConstants.FACET_CATS, PortletSession.APPLICATION_SCOPE));

			// dispatch to jsp
			normalView.include(req, res);
		}
	}

	/**
	 * handle the incoming event
	 */
	@Override
	public void processEvent(EventRequest req, EventResponse res) throws PortletException, IOException {
		// checking event
		if (req.getEvent() == null) {
			// dispatch to FacetPortletConstants.ERROR jsp
			errorView.include(req, res);
		}

		String reaction = getReaction(req.getEvent().getQName()).getLocalPart();
		logger.info("Event received " + req.getEvent() + " - launching reaction : " + reaction);

		if (FacetPortletConstants.EVENT_DISPLAY_RESULTS.equals(reaction)) {

			if (!(req.getEvent().getValue() instanceof ResultSet)) {
				// dispatch to FacetPortletConstants.ERROR jsp
				errorView.include(req, res);
			}

			ResultSet receveidResultSet = (ResultSet) req.getEvent().getValue();
			logger.info("Event [" + FacetPortletConstants.EVENT_DISPLAY_RESULTS + "] received with ResultSet [" + receveidResultSet + "].");
			try {
				FacetBundle facets = processResultSet(receveidResultSet);
				// adding facets map in user session for later use
				req.getPortletSession().setAttribute(FacetPortletConstants.FACET_CATS, facets, PortletSession.APPLICATION_SCOPE);
			} catch (WebLabCheckedException e) {
				logger.debug("No facets.");
				// no facets here, reset the current ones
				req.getPortletSession().removeAttribute(FacetPortletConstants.FACET_CATS);
			}

		} else if (FacetPortletConstants.EVENT_RESET_FILTER.equals(reaction)) {
			logger.info("Event [" + FacetPortletConstants.EVENT_RESET_FILTER + "] received. Resetting facets.");
			req.getPortletSession().removeAttribute(FacetPortletConstants.FACET_CATS);
		} else {
			logger.warn("This kind of event wasn't planned [" + getReaction(req.getEvent().getQName()).getLocalPart() + "]. Ignoring it.");
		}
	}

	@Override
	public void processAction(ActionRequest req, ActionResponse res) throws PortletException, IOException {
		String facetQueryURI = req.getParameter(FacetPortletConstants.FACET_QUERY_URI);

		logger.info("Action received. Launching query [" + facetQueryURI + "].");

		if (req.getPortletSession().getAttribute(FacetPortletConstants.FACET_CATS, PortletSession.APPLICATION_SCOPE) == null) {
			logger.error("No facets in session.");
			// no facets here that's a problem man
			req.setAttribute(FacetPortletConstants.ERROR, "No facets in session.");
		}

		Object ofacet = req.getPortletSession().getAttribute(FacetPortletConstants.FACET_CATS, PortletSession.APPLICATION_SCOPE);
		if (!(ofacet instanceof FacetBundle)) {
			logger.error("Facets in session are not correct.");
			// no facets here that's a problem man
			req.setAttribute(FacetPortletConstants.ERROR, "Facets in session are not correct.");
		}

		FacetBundle facets = (FacetBundle) ofacet;

		if (!facets.getFQueryByURIMap().containsKey(facetQueryURI)) {
			logger.error("Facet query [" + facetQueryURI + "] is not session.");
			// no facets here that's a problem man
			req.setAttribute(FacetPortletConstants.ERROR, "Facet query [" + facetQueryURI + "] is not session.");
		}

		Query facetQueryToLaunch = facets.getFQueryByURIMap().get(facetQueryURI);

		// remove annotation that tell the query came from facet suggestion
		// (it's now a query selected by user)
		facetQueryToLaunch = prepareQuery(facetQueryToLaunch);

		// TODO launch the right event to search portlet
		// create args for search service
		SearchArgs args = new SearchArgs();
		args.setLimit(10);
		args.setOffset(0);
		args.setQuery(facetQueryToLaunch);

		// launch event to broadcast the searchArgs to all portlets
		// (searchportlet should catch the event itself)
		QName qnOfEvent = QName.valueOf(this.getPortletConfig().getInitParameter(PORTLET_ACTION_NAMESPACE + FacetPortletConstants.ACTION_PUBLISH_QUERY));
		logger.info("publishing searchArgs in event : [" + qnOfEvent + "]");
		res.setEvent(qnOfEvent, args);
	}

	protected Query prepareQuery(Query q) {
		JenaPoKHelper hlpr;
		if (q.getAnnotation().size() > 0) {
			for(Annotation a : q.getAnnotation()){
				hlpr = new JenaPoKHelper(a);
				hlpr.removeLitStat(q.getUri(), WebLabRetrieval.IS_EXPRESSED_WITH, FACET_SUGGESTION);
				hlpr.commit();
			}
		}
		for (Query query : ResourceUtil.getSelectedSubResources(q, Query.class)) {
			query = prepareQuery(query);
		}
		return q;
	}

	/**
	 * Processing a ResultSet to extract the facet and related information and
	 * produce a valid FacetBundle.
	 * 
	 * @param receveidResultSet
	 * @return
	 * @throws WebLabCheckedException
	 */
	protected FacetBundle processResultSet(ResultSet receveidResultSet) throws WebLabCheckedException {
		// getting facet queries
		JenaResourceHelper hlpr = new JenaResourceHelper(receveidResultSet);

		Set<Query> facetQueries = filterResources(receveidResultSet, hlpr.getSubjsOnPredLit(WebLabRetrieval.IS_EXPRESSED_WITH, FACET_SUGGESTION), Query.class);

		FacetBundle bundle = new FacetBundle();
		try {
			if (facetQueries.size() == 0) {
				throw new WebLabCheckedException("No facet queries received in last resultSet. Cleaning old ones.");
			}
			logger.info(facetQueries.size() + " facet queries received.");

			for (Query q : facetQueries) {
				if (!(q instanceof ComposedQuery)) {
					throw new WebLabCheckedException("One of the facet query [" + q.getUri() + "] received isn't a composedQuery. We will ignor it.");
				} else {
					ComposedQuery facetQuery = (ComposedQuery) q;
					// check the link to original query
					List<String> origin = hlpr.getRessOnPredSubj(facetQuery.getUri(), WebLabRetrieval.IS_LINKED_TO);
					if (origin == null || origin.size() != 1) {
						throw new WebLabCheckedException("One of the facet query [" + q.getUri()
								+ "] received isn't correclty linked to the orginial query. We will ignor it.");
					}
					// get count
					List<String> counts = hlpr.getLitsOnPredSubj(facetQuery.getUri(), WebLabRetrieval.HAS_NUMBER_OF_RESULTS);
					if (counts == null || counts.size() != 1) {
						throw new WebLabCheckedException("One of the facet query [" + q.getUri()
								+ "] received isn't correclty built upon the orginial query. We will ignor it.");
					}
					int count = Integer.parseInt(counts.get(0));
					// check that the facet is built upon original query
					Set<Query> originalQuery = filterResources(receveidResultSet, new HashSet<String>(origin), Query.class);
					if (originalQuery == null || originalQuery.size() < 1) {
						throw new WebLabCheckedException("One of the facet query [" + q.getUri()
								+ "] received isn't correclty built upon the orginial query. We will ignor it.");
					}
					// process and extract facet query by getting scope and
					// label
					// ugly : suppose the facet query is built by solr facet
					// suggestion : the filter query is the second one
					Query filter = facetQuery.getQuery().get(1);
					if (!(filter instanceof StringQuery)) {
						throw new WebLabCheckedException("One of the facet query [" + q.getUri() + "] received isn't isn't a StringQuery but a ["
								+ filter.getClass().getSimpleName() + "]. We will ignor it.");
					}
					StringQuery sfilter = (StringQuery) filter;
					List<String> scopes = hlpr.getRessOnPredSubj(sfilter.getUri(), HAS_SCOPE);
					if (scopes == null || scopes.size() != 1) {
						throw new WebLabCheckedException("One of the facet query [" + q.getUri()
								+ "] received doesn't have a correct unique scope. We will ignor it.");
					}

					String scope = scopes.get(0);
					Facet facet = new Facet(sfilter.getRequest(), count, facetQuery);
					if (!bundle.containsKey(scope)) {
						bundle.put(scope, new FacetCategory(scope, portletConf.getFacetLabelKey().get(scope), facet));
					} else {
						bundle.get(scope).getFacets().add(facet);
					}
				}
			}
			if (bundle.isEmpty()) {
				throw new WebLabCheckedException("Not any facet to present...");
			} else {
				// ordering facet for presentation
				for (FacetCategory cat : bundle.values()) {
					Collections.sort(cat.getFacets(), Facet.FacetComparator.getInstance());
				}
			}

		} catch (WebLabCheckedException e) {
			logger.warn(e.getMessage());
			throw new WebLabCheckedException("No facet to present.", e);
		}
		return bundle;
	}

	protected <T extends Resource> Set<T> filterResources(Resource receveidResultSet, Set<String> acceptedURIs, Class<T> resClass) {
		Set<T> filteredQueries = new HashSet<T>();
		List<T> queries = ResourceUtil.getSelectedSubResources(receveidResultSet, resClass);
		for (T q : queries) {
			if (acceptedURIs.contains(q.getUri())) {
				filteredQueries.add(q);
			}
		}
		return filteredQueries;
	}

	public FacetConfig getPortletConf() {
		return portletConf;
	}

	public void setPortletConf(FacetConfig portletConf) {
		this.portletConf = portletConf;
	}
}
